/**********************************************************************
Copyright (c) 2010 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.jpa;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;
import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.MappedSuperclass;

/**
 * Series of utility methods for use with annotation processing and persistence.
 */
public class APTUtils
{
    // TODO Tie this in to TypeManager in core
    private static Set<String> LIST_CLASSNAMES = null;
    private static Set<String> SET_CLASSNAMES = null;
    private static Set<String> MAP_CLASSNAMES = null;
    private static Set<String> COLLECTION_CLASSNAMES = null;
    static
    {
        LIST_CLASSNAMES = new HashSet<String>();
        LIST_CLASSNAMES.add("java.util.List");
        LIST_CLASSNAMES.add("java.util.ArrayList");
        LIST_CLASSNAMES.add("java.util.AbstractList");
        LIST_CLASSNAMES.add("java.util.Stack");
        LIST_CLASSNAMES.add("java.util.Vector");
        LIST_CLASSNAMES.add("java.util.LinkedList");
        SET_CLASSNAMES = new HashSet<String>();
        SET_CLASSNAMES.add("java.util.Set");
        SET_CLASSNAMES.add("java.util.HashSet");
        SET_CLASSNAMES.add("java.util.AbstractSet");
        SET_CLASSNAMES.add("java.util.LinkedHashSet");
        SET_CLASSNAMES.add("java.util.TreeSet");
        SET_CLASSNAMES.add("java.util.SortedSet");
        MAP_CLASSNAMES = new HashSet<String>();
        MAP_CLASSNAMES.add("java.util.Map");
        MAP_CLASSNAMES.add("java.util.HashMap");
        MAP_CLASSNAMES.add("java.util.AbstractMap");
        MAP_CLASSNAMES.add("java.util.Hashtable");
        MAP_CLASSNAMES.add("java.util.LinkedHashMap");
        SET_CLASSNAMES.add("java.util.TreeMap");
        MAP_CLASSNAMES.add("java.util.SortedMap");
        MAP_CLASSNAMES.add("java.util.Properties");
        COLLECTION_CLASSNAMES = new HashSet<String>();
        COLLECTION_CLASSNAMES.add("java.util.Collection");
        COLLECTION_CLASSNAMES.add("java.util.AbstractCollection");
        COLLECTION_CLASSNAMES.add("java.util.Queue");
        COLLECTION_CLASSNAMES.add("java.util.PriorityQueue");
    }

    /**
     * Method to return the JPA2 type category for a type.
     * @param typeName The type name (e.g java.lang.String, java.util.Collection)
     * @return The type category
     */
    public static TypeCategory getTypeCategoryForTypeMirror(String typeName)
    {
        if (COLLECTION_CLASSNAMES.contains(typeName))
        {
            return TypeCategory.COLLECTION;
        }
        else if (SET_CLASSNAMES.contains(typeName))
        {
            return TypeCategory.SET;
        }
        else if (LIST_CLASSNAMES.contains(typeName))
        {
            return TypeCategory.LIST;
        }
        else if (MAP_CLASSNAMES.contains(typeName))
        {
            return TypeCategory.MAP;
        }
        return TypeCategory.ATTRIBUTE;
    }

    /**
     * Convenience accessor for all field members of the supplied type element.
     * @param el The type element
     * @return The field members
     */
    public static List<? extends Element> getFieldMembers(TypeElement el)
    {
        List<? extends Element> members = el.getEnclosedElements();
        List<Element> fieldMembers = new ArrayList<Element>();
        Iterator<? extends Element> memberIter = members.iterator();
        while (memberIter.hasNext())
        {
            Element member = memberIter.next();
            if (member.getKind() == ElementKind.FIELD)
            {
                fieldMembers.add(member);
            }
        }
        return fieldMembers;
    }

    /**
     * Convenience accessor for all property members of the supplied type element.
     * @param el The type element
     * @return The property members
     */
    public static List<? extends Element> getPropertyMembers(TypeElement el)
    {
        List<? extends Element> members = el.getEnclosedElements();
        List<Element> propertyMembers = new ArrayList<Element>();
        Iterator<? extends Element> memberIter = members.iterator();
        while (memberIter.hasNext())
        {
            Element member = memberIter.next();
            if (member.getKind() == ElementKind.METHOD)
            {
                ExecutableElement method = (ExecutableElement)member;
                if (isJavaBeanGetter(method) || isJavaBeanSetter(method))
                {
                    propertyMembers.add(member);
                }
            }
        }
        return propertyMembers;
    }

    /**
     * Convenience accessor for members for the default access type of the supplied type element.
     * If properties are annotated then returns all properties, otherwise returns all fields. 
     * @param el The type element
     * @return The members
     */
    public static List<? extends Element> getDefaultAccessMembers(TypeElement el)
    {
        Iterator<? extends Element> memberIter = el.getEnclosedElements().iterator();
        while (memberIter.hasNext())
        {
            Element member = memberIter.next();
            
            if (isMethod(member))
            {
                ExecutableElement method = (ExecutableElement)member;
                if (isJavaBeanGetter(method) || isJavaBeanSetter(method))
                {
                    // Property
                    Iterator<? extends AnnotationMirror> annIter = member.getAnnotationMirrors().iterator();
                    while (annIter.hasNext())
                    {
                        AnnotationMirror ann = annIter.next();
                        String annTypeName = ann.getAnnotationType().toString();
                        if (annTypeName.startsWith("javax.persistence"))
                        {
                            return getPropertyMembers(el);
                        }
                    }
                }
            }
        }
        return getFieldMembers(el);
    }

    /**
     * Convenience method to return if this class element has any of the defining JPA annotations.
     * @param el The class element
     * @return Whether it is to be considered a JPA annotated class
     */
    public static boolean isJPAAnnotated(TypeElement el)
    {
        if ((el.getAnnotation(Entity.class) != null) ||
            (el.getAnnotation(MappedSuperclass.class) != null) ||
            (el.getAnnotation(Embeddable.class) != null))
        {
            return true;
        }
        return false;
    }

    public static Object getValueForAnnotationAttribute(Element elem, Class annotCls, String attribute)
    {
        List<? extends AnnotationMirror> anns = elem.getAnnotationMirrors();
        Iterator<? extends AnnotationMirror> annIter = anns.iterator();
        while (annIter.hasNext())
        {
            AnnotationMirror ann = annIter.next();
            if (ann.getAnnotationType().toString().equals(annotCls.getName()))
            {
                Map<? extends ExecutableElement, ? extends AnnotationValue> values = ann.getElementValues();
                for (ExecutableElement ex : values.keySet())
                {
                    if (ex.getSimpleName().toString().equals(attribute))
                    {
                        return values.get(ex).getValue();
                    }
                }
            }
        }
        return null;
    }

    public static TypeMirror getDeclaredType(Element elem)
    {
        if (elem.getKind() == ElementKind.FIELD)
        {
            return elem.asType();
        }
        else if (elem.getKind() == ElementKind.METHOD)
        {
            return ((ExecutableElement) elem).getReturnType();
        }
        else
        {
            throw new IllegalArgumentException("Unable to get type for " + elem);
        }
    }

    /**
     * Convenience method to return if the provided element represents a method (otherwise a field).
     * @param elem The element
     * @return Whether it represents a method
     */
    public static boolean isMethod(Element elem)
    {
        return elem != null && ExecutableElement.class.isInstance(elem) && elem.getKind() == ElementKind.METHOD;
    }

    /**
     * Convenience method returning if the provided method returns void.
     * @param method The method
     * @return Whether it returns void
     */
    public static boolean returnsVoid(ExecutableElement method)
    {
        TypeMirror type = method.getReturnType();
        return (type != null && type.getKind() == TypeKind.VOID);
    }

    /**
     * Convenience method returning if the provided method returns boolean.
     * @param method The method
     * @return Whether it returns boolean
     */
    public static boolean returnsBoolean(ExecutableElement method)
    {
        TypeMirror type = method.getReturnType();
        return (type != null && (type.getKind() == TypeKind.BOOLEAN  || "java.lang.Boolean".equals(type.toString())));
    }

    /**
     * Convenience method to return if the provided method is a java bean getter.
     * @param method The method
     * @return Whether it is a java bean getter
     */
    public static boolean isJavaBeanGetter(ExecutableElement method)
    {
        String methodName = method.getSimpleName().toString();
        if (method.getKind() == ElementKind.METHOD && method.getParameters().isEmpty())
        {
            if (returnsBoolean(method) && methodName.startsWith("is"))
            {
                return true;
            }
            else if (methodName.startsWith("get") && !returnsVoid(method))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Convenience method to return if the provided method is a java bean setter.
     * @param method The method
     * @return Whether it is a java bean setter
     */
    public static boolean isJavaBeanSetter(ExecutableElement method)
    {
        String methodName = method.getSimpleName().toString();
        return method.getKind() == ElementKind.METHOD && methodName.startsWith("set") &&
            method.getParameters().isEmpty() && !returnsVoid(method);
    }

    public static String getPersistentName(Element el)
    {
        if (el.getKind() == ElementKind.FIELD)
        {
            return el.toString();
        }
        else if (el.getKind() == ElementKind.METHOD)
        {
            ExecutableElement method = (ExecutableElement) el;
            if (isJavaBeanGetter(method) || isJavaBeanSetter(method))
            {
                String name = method.getSimpleName().toString();
                if (name.startsWith("is"))
                {
                    name = name.substring(2);
                }
                else
                {
                    name = name.substring(3);
                }
                return Character.toLowerCase(name.charAt(0)) + name.substring(1);
            }
        }
        return null;
    }

    public static enum TypeCategory
    {
        COLLECTION("CollectionAttribute"),
        SET("SetAttribute"),
        LIST("ListAttribute"),
        MAP("MapAttribute"),
        ATTRIBUTE("SingularAttribute");

        private String type;

        private TypeCategory(String type)
        {
            this.type = type;
        }

        public String getTypeName()
        {
            return type;
        }
    }
}