/**********************************************************************
Copyright (c) 2010 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.jpa;

import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.ArrayType;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.PrimitiveType;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.Elements;
import javax.lang.model.util.Types;
import javax.persistence.AccessType;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.Transient;
import javax.tools.JavaFileObject;

import org.datanucleus.jpa.APTUtils.TypeCategory;

/**
 * Annotation processor for JPA to generate "dummy" classes for all Entity classes for use with Criteria API.
 * Any class ({MyClass}) that has a JPA "class" annotation will have a stub class ({MyClass}_) generated.
 * <ul>
 * <li>For each managed class X in package p, a metamodel class X_ in package p is created.</li>
 * <li>The name of the metamodel class is derived from the name of the managed class by appending "_" 
 * to the name of the managed class.</li>
 * <li>The metamodel class X_ must be annotated with the javax.persistence.StaticMetamodel annotation</li>
 * <li>If class X extends another class S, where S is the most derived managed class (i.e., entity or 
 * mapped superclass) extended by X, then class X_ must extend class S_, where S_ is the meta-model class 
 * created for S.</li>
 * <li>For every persistent non-collection-valued attribute y declared by class X, where the type of y is Y, 
 * the metamodel class must contain a declaration as follows:
 * <pre>public static volatile SingularAttribute<X, Y> y;</pre></li>
 * <li>For every persistent collection-valued attribute z declared by class X, where the element type
 * of z is Z, the metamodel class must contain a declaration as follows:
 *     <ul>
 *     <li>if the collection type of z is java.util.Collection, then 
 *     <pre>public static volatile CollectionAttribute<X, Z> z;</pre></li>
 *     <li>if the collection type of z is java.util.Set, then
 *     <pre>public static volatile SetAttribute<X, Z> z;</pre></li>
 *     <li>if the collection type of z is java.util.List, then
 *     <pre>public static volatile ListAttribute<X, Z> z;</pre></li>
 *     <li>if the collection type of z is java.util.Map, then
 *     <pre>public static volatile MapAttribute<X, K, Z> z;</pre>
 *     where K is the type of the key of the map in class X</li>
 *     </ul>
 * </li>
 * </ul>
 */
@SupportedAnnotationTypes({"javax.persistence.Entity", "javax.persistence.Embeddable", "javax.persistence.MappedSuperclass"})
@SupportedSourceVersion(SourceVersion.RELEASE_6)
public class JPACriteriaProcessor extends AbstractProcessor
{
    String suffix = "_";

    Types typesHandler;

    protected static Class[] annotationsWithTargetEntity =
        new Class[] {OneToOne.class, OneToMany.class, ManyToOne.class, ManyToMany.class};

    /* (non-Javadoc)
     * @see javax.annotation.processing.AbstractProcessor#process(java.util.Set, javax.annotation.processing.RoundEnvironment)
     */
    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv)
    {
        if (roundEnv.processingOver())
        {
            return false;
        }

        typesHandler = processingEnv.getTypeUtils();
        Set<? extends Element> elements = roundEnv.getRootElements();
        for (Element e : elements)
        {
            processClass((TypeElement)e);
        }
        return false;
    }

    /**
     * Handler for processing a JPA annotated class to create the criteria class stub.
     * @param el The class element
     */
    protected void processClass(TypeElement el)
    {
        if (el == null || !APTUtils.isJPAAnnotated(el))
        {
            return;
        }

        // TODO Set imports to only include the classes we require
        // TODO Set references to other classes to be the class name and put the package in the imports
        // TODO Support specification of the location for writing the class source files
        Elements elementUtils = processingEnv.getElementUtils();
        String className = elementUtils.getBinaryName(el).toString();
        String pkgName = className.substring(0, className.lastIndexOf('.'));
        String classSimpleName = className.substring(className.lastIndexOf('.') + 1);
        String classNameNew = className + suffix;
        System.out.println("JPA2 CriteriaProcessor orig=" + className + " new=" + classNameNew);

        TypeElement superEl = getPersistentSupertype(el);
        try
        {
            JavaFileObject javaFile = processingEnv.getFiler().createSourceFile(classNameNew);
            Writer w = javaFile.openWriter();
            try
            {
                w.append("package " + pkgName + ";\n");
                w.append("\n");
                w.append("import javax.persistence.metamodel.*;\n");
                w.append("\n");
                w.append("@StaticMetamodel(" + classSimpleName + ".class)\n");
                w.append("public class " + classSimpleName + suffix);
                if (superEl != null)
                {
                    String superClassName = elementUtils.getBinaryName(superEl).toString();
                    w.append(" extends ").append(superClassName + suffix);
                }
                w.append("\n");
                w.append("{\n");

                // Find the members to use for persistence processing
                AccessType clsAccessType = (AccessType) APTUtils.getValueForAnnotationAttribute(el, AccessType.class, "value");
                List<? extends Element> members = null;
                if (clsAccessType == AccessType.FIELD)
                {
                    // Only use fields
                    members = APTUtils.getFieldMembers(el);
                }
                else if (clsAccessType == AccessType.PROPERTY)
                {
                    // Only use properties
                    members = APTUtils.getPropertyMembers(el);
                }
                else
                {
                    // Default access type - whichever type (field or method) is annotated first
                    members = APTUtils.getDefaultAccessMembers(el);
                }

                if (members != null)
                {
                    Iterator<? extends Element> iter = members.iterator();
                    while (iter.hasNext())
                    {
                        Element member = iter.next();
                        boolean isTransient = false;
                        List<? extends AnnotationMirror> annots = member.getAnnotationMirrors();
                        if (annots != null)
                        {
                            Iterator<? extends AnnotationMirror> annotIter = annots.iterator();
                            while (annotIter.hasNext())
                            {
                                AnnotationMirror annot = annotIter.next();
                                if (annot.getAnnotationType().toString().equals(Transient.class.getName()))
                                {
                                    // Ignore this
                                    isTransient = true;
                                    break;
                                }
                            }
                        }

                        if (!isTransient)
                        {
                            if (member.getKind() == ElementKind.FIELD ||
                                (member.getKind() == ElementKind.METHOD && APTUtils.isJavaBeanGetter((ExecutableElement) member)))
                            {
                                TypeMirror type = APTUtils.getDeclaredType(member);
                                String typeName = getDeclaredTypeName(type, true);
                                TypeCategory cat = APTUtils.getTypeCategoryForTypeMirror(typeName);
                                String memberName = APTUtils.getPersistentName(member);

                                w.append("    public static volatile " + cat.getTypeName());
                                w.append("<" + classSimpleName + ", ");

                                if (cat == TypeCategory.ATTRIBUTE)
                                {
                                    w.append(type.toString());
                                }
                                else if (cat == TypeCategory.MAP)
                                {
                                    TypeMirror keyType = getTypeParameter(member, type, 0, false);
                                    String keyTypeName = getDeclaredTypeName(keyType, true);
                                    TypeMirror valueType = getTypeParameter(member, type, 1, true);
                                    String valueTypeName = getDeclaredTypeName(valueType, true);
                                    w.append(keyTypeName + ", " + valueTypeName);
                                }
                                else
                                {
                                    TypeMirror elementType = getTypeParameter(member, type, 0, true);
                                    String elementTypeName = getDeclaredTypeName(elementType, true);
                                    w.append(elementTypeName);
                                }
                                w.append("> " + memberName + ";\n");
                            }
                        }
                    }
                }

                w.append("}\n");
                w.flush();
            }
            finally
            {
                w.close();
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    String getDeclaredTypeName(TypeMirror type, boolean box)
    {
        if (type == null || type.getKind() == TypeKind.NULL || type.getKind() == TypeKind.WILDCARD)
        {
            return "java.lang.Object";
        }
        if (type.getKind() == TypeKind.ARRAY)
        {
            TypeMirror comp = ((ArrayType)type).getComponentType();
            return getDeclaredTypeName(comp, false);
        }

        if (box && isPrimitive(type))
        {
            type = processingEnv.getTypeUtils().boxedClass((PrimitiveType)type).asType();
        }
        if (isPrimitive(type))
        {
            return ((PrimitiveType)type).toString();
        }
        return typesHandler.asElement(type).toString();
    }

    private static boolean isPrimitive(TypeMirror type)
    {
        TypeKind kind = type.getKind();
        return kind == TypeKind.BOOLEAN || kind == TypeKind.BYTE || kind == TypeKind.CHAR || 
            kind == TypeKind.DOUBLE || kind == TypeKind.FLOAT || kind == TypeKind.INT || 
            kind == TypeKind.LONG || kind == TypeKind.SHORT;
    }

    TypeMirror getTypeParameter(Element element, TypeMirror type, int position, boolean checkTarget)
    {
        if (type.getKind() == TypeKind.ARRAY)
        {
            return ((ArrayType)type).getComponentType();
        }
        if (type.getKind() != TypeKind.DECLARED)
        {
            return null;
        }

        if (checkTarget) 
        {
            // Check annotations for a value
            TypeMirror target = null;
            for (int i=0;i<annotationsWithTargetEntity.length;i++)
            {
                Object targetValue = APTUtils.getValueForAnnotationAttribute(element, annotationsWithTargetEntity[i], "targetEntity");
                if (targetValue != null)
                {
                    target = (TypeMirror)targetValue;
                    break;
                }
            }
            if (target != null)
            {
                return target;
            }
        }

        // Use generics
        List<? extends TypeMirror> params = ((DeclaredType)type).getTypeArguments();
        TypeMirror param = (params == null || params.size() < position+1) ? typesHandler.getNullType() : params.get(position);
        return param;
    }

    /**
     * Method to find the next persistent supertype above this one.
     * @param element The element
     * @return Its next parent that is persistable (or null if no persistable predecessors)
     */
    public TypeElement getPersistentSupertype(TypeElement element)
    {
        TypeMirror superType = element.getSuperclass();
        if (superType == null || (element != null && "java.lang.Object".equals(element.toString())))
        {
            return null;
        }

        TypeElement superElement = (TypeElement) processingEnv.getTypeUtils().asElement(superType);
        if (APTUtils.isJPAAnnotated(superElement))
        {
            return superElement;
        }
        return getPersistentSupertype(superElement);
    }
}